
#include "RealmSceneElevationSubMap.h"
#include "RealmSceneElevationMapRepository.h"

#include <time.h>

RealmSceneElevationSubMap::RealmSceneElevationSubMap(
		RealmSceneElevationMap* parent,
		RealmSceneElevationSubMap* centermap,
		RealmMapKey mapkey,
		RealmSceneElevationMapRepository* maprepo
	) :
	RealmSceneElevationMap(maprepo, mapkey, SCENE_SPIRAL_SIZE)

{
	ParentMap = parent;
	if (centermap != nullptr)
		CenterMap = centermap;
	else
		CenterMap = this;
	CurrentPhase = 0;

	HexSpiralCoordinate centercoord( mapkey.MapCenterKey.getSpiralIndex());
	HexSpiralCoordinate offsetcoord( mapkey.CenterSpiralOffset);

	SpiralIndexInParentMap = (centercoord + offsetcoord).getHexIndex();


	// the seeds are derived from the centermap, while elevation center coords come from parent map
	// if (MapKey.CenterSpiralOffset == 0) {
		// center map
		// if this is the centermap, seed will be derived directly from parent map
		// NumGenerator.init(ParentMap->getSubSeed(SpiralIndexInParentMap));
	// } else {
		// use centermap + mapkey spiral offset to get our seed from centermap.
		// NumGenerator.init(CenterMap->getChildSeed(MapKey.CenterSpiralOffset));
	// }

	// todo replace with deterministic map hash that works better
	NumGenerator.init(time(NULL));

	initTriangles(0);
}

bool RealmSceneElevationSubMap::isInverted() {
	return MapKey.MapCenterKey.displaylevel == 2;
}

int RealmSceneElevationSubMap::findElevation(long int spiralidx) {
	if (spiralidx >= MapSize) {

		HexSpiralCoordinate loc(spiralidx);

		HexSpiralCoordinate parloc = loc.getParentSpiralCoordinate(!MapKey.isInverted());

		RealmMapKey translated_key(MapKey.MapCenterKey, parloc.getHexIndex() );

		RealmSceneElevationMap* translated_map = MapRepository->findSceneElevationMap(translated_key);
		translated_map->populate();

		HexSpiralCoordinate translated_map_center = parloc.getChildSpiralCoordinate(MapKey.isInverted());
		HexSpiralCoordinate translated_map_coord = loc - translated_map_center;

		return translated_map->getElevation(translated_map_coord.getHexIndex());

	} else {
		return getElevation(spiralidx);
	}
}


void RealmSceneElevationSubMap::initNeighbors() {

	HexSpiralCoordinate spiralcoord(SpiralIndexInParentMap );
	for (int dir = 0; dir < 6; dir++) {
		int realdir = dir;
		if (isInverted())
			realdir = dir > 0 ? dir - 1 : 5;
		HexSpiralCoordinate neighborcoord = spiralcoord.calcMovement(realdir);

		RealmMapKey NeighborKey = MapKey;
		HexSpiralCoordinate absmapcoord = HexSpiralCoordinate(MapKey.MapCenterKey.getSpiralIndex());
		NeighborKey.CenterSpiralOffset = (neighborcoord - absmapcoord).getHexIndex();

		NeighborMap[dir] = MapRepository->findSceneElevationSubMap(NeighborKey);
	}
}

void RealmSceneElevationSubMap::initTriangles(int phase) {
	// note: initNeighbors() should be called BEFORE initTriangles() so the neighbor side triangle relations can be setup correctly
	for (int i = 0; i < 6; i++) {
		if (phase == 0) {
			ElevationTriangle[i].init(this, i, &NumGenerator);
		} else {
			int previdx = (i == 0 ? 5 : i - 1),
				nextidx = (i == 5 ? 0 : i + 1);

			int neighbor_triangle_index = i+3 < 6 ? i+3 : i - 3;

			ElevationTriangle[i].setNeighborSides(
						ElevationTriangle[previdx].getTriangleSide(2),
						NeighborMap[i]->getTriangle( neighbor_triangle_index )->getTriangleSide(1),
						ElevationTriangle[nextidx].getTriangleSide(0)
			);
		}
	}
}

int RealmSceneElevationSubMap::getBaseVariance() {
	int variance;
	// just some test values to play with that "look" reasonably good for now
	switch (MapKey.MapCenterKey.displaylevel) {
		case 1: variance = 500; break;
		case 2: variance = 700; break;
		case 3: variance = 1200; break;
	}

	return variance;
}

int RealmSceneElevationSubMap::getCornerIndex(int n) {
	return (n == 0) ?
		546 :
		468 + n * 13;
}

int RealmSceneElevationSubMap::getCorner(int n) {
	int ei = getCornerIndex(n);
	return getElevation(ei);
}


void RealmSceneElevationSubMap::populateBatch(int phase) {

	int t;

	// phase 0 sets up neighbor triangle init
	if (phase == 0) {
		// post-init setup for establishing neighbor object relations and triangles
		// populates triangle corners with initial values based on neighbor hex height values
		initNeighbors();
		initTriangles(1);

		HexSpiralCoordinate here(SpiralIndexInParentMap);
		long int neighbor_spiral_index[6];
		int tcorner[6];
		for (t = 0; t < 6; t++) {
			int realdir = t;
			if (isInverted()) {
				realdir = t > 0 ? t - 1 : 5;
			}
			HexSpiralCoordinate neighbor_coord = here.calcMovement(realdir);
			neighbor_spiral_index[t] = neighbor_coord.getHexIndex();
		}

		// find value for each hex corner, populate all applicable triangles
		for ( t = 0; t < 6; t++) {

			int invt = ( t + 3 > 5 ? t - 3 : t + 3);
			int prevt = t > 0 ? t - 1 : 5;
			int nextinvt = invt < 5 ? invt + 1 : 0;
			int previnvt = invt > 0 ? invt - 1 : 5;

			// check neighbor corners first.  if already populated, use existing values instead.
			if (getCorner(t) != -1)
				tcorner[t] = getCorner(t);
			else if (NeighborMap[t]->getCorner(nextinvt) != -1)
				tcorner[t] = NeighborMap[t]->getCorner(nextinvt);
			else if (NeighborMap[prevt]->getCorner(previnvt) != -1)
				tcorner[t] = NeighborMap[prevt]->getCorner(previnvt);
			else {
				int inputvals[3];
				inputvals[0] = ParentMap->findElevation(SpiralIndexInParentMap);

				inputvals[1] = ParentMap->findElevation(neighbor_spiral_index[prevt]);
				inputvals[2] = ParentMap->findElevation(neighbor_spiral_index[t]);
				tcorner[t] = NumGenerator.calculateVarianceValue( getBaseVariance() * 7 /* 13 */ , inputvals, 3);
			}

			ElevationTriangle[t].getTriangleSide(0)->setSpace(4, 13, tcorner[t], false);

			NeighborMap[t]->getTriangle(invt)->getTriangleSide(1)->setSpace(4, 13, tcorner[t], false);
			NeighborMap[prevt]->getTriangle(previnvt)->getTriangleSide(0)->setSpace(4, 13, tcorner[t], false);
		}
		int centerheight = NumGenerator.calculateVarianceValue( getBaseVariance() * 7 /* 13 */, tcorner, 6);
		setElevation(0, centerheight);

	} else {
		for (t = 0; t < 6; t++) {
			// initialize neighbor maps before populating triangles (phase 1 or later)
			if (phase == 1)
				NeighborMap[t]->populateToPhase(0); // phase - 1); // was: 0

			ElevationTriangle[t].populateToPhase(phase);
		}
	}
}

void RealmSceneElevationSubMap::populateToPhase( int phase) {

	ParentMap->populate();
	while (CurrentPhase <= phase) {
		populateBatch(CurrentPhase++);
	}
}

void RealmSceneElevationSubMap::populate() {
	populateToPhase(8);
}

void RealmSceneElevationSubMap::populateStep() {
	populateToPhase(CurrentPhase);
}

void RealmSceneElevationSubMap::populateNeighborStep() {
	populateStep();
	for (int t = 0; t < 6; t++) {
		NeighborMap[t]->populateStep();
	}
}

void RealmSceneElevationSubMap::populateNeighbors() {
	populateStep();
	for (int t = 0; t < 6; t++) {
		NeighborMap[t]->populate();
	}
}

